@@ -7,30 +7,30 @@ RETRY_TIMES - how many times to retry a failed page
 RETRY_HTTP_CODES - which HTTP response codes to retry
 
 Failed pages are collected on the scraping process and rescheduled at the end,
-once the spider has finished crawling all regular (non failed) pages.
+once the spider has finished crawling all regular (non-failed) pages.
 """
+
+from __future__ import annotations
+
 from logging import Logger, getLogger
-from typing import Optional, Union
+from typing import TYPE_CHECKING
 
-from twisted.internet import defer
-from twisted.internet.error import (
-    ConnectError,
-    ConnectionDone,
-    ConnectionLost,
-    ConnectionRefusedError,
-    DNSLookupError,
-    TCPTimedOutError,
-    TimeoutError,
-)
-from twisted.web.client import ResponseFailed
-
-from scrapy.core.downloader.handlers.http11 import TunnelError
 from scrapy.exceptions import NotConfigured
-from scrapy.http.request import Request
-from scrapy.spiders import Spider
+from scrapy.utils.misc import load_object
 from scrapy.utils.python import global_object_name
 from scrapy.utils.response import response_status_message
 
+if TYPE_CHECKING:
+    # typing.Self requires Python 3.11
+    from typing_extensions import Self
+
+    from scrapy.crawler import Crawler
+    from scrapy.http import Response
+    from scrapy.http.request import Request
+    from scrapy.settings import BaseSettings
+    from scrapy.spiders import Spider
+
+
 retry_logger = getLogger(__name__)
 
 
@@ -38,12 +38,12 @@ def get_retry_request(
     request: Request,
     *,
     spider: Spider,
-    reason: Union[str, Exception] = "unspecified",
-    max_retry_times: Optional[int] = None,
-    priority_adjust: Optional[int] = None,
+    reason: str | Exception | type[Exception] = "unspecified",
+    max_retry_times: int | None = None,
+    priority_adjust: int | None = None,
     logger: Logger = retry_logger,
     stats_base_key: str = "retry",
-):
+) -> Request | None:
     """
     Returns a new :class:`~scrapy.Request` object to retry the specified
     request, or ``None`` if retries of the specified request have been
@@ -85,6 +85,7 @@ def get_retry_request(
     retry-related job stats
     """
     settings = spider.crawler.settings
+    assert spider.crawler.stats
     stats = spider.crawler.stats
     retry_times = request.meta.get("retry_times", 0) + 1
     if max_retry_times is None:
@@ -122,36 +123,24 @@ def get_retry_request(
 
 
 class RetryMiddleware:
-    # IOError is raised by the HttpCompression middleware when trying to
-    # decompress an empty response
-    EXCEPTIONS_TO_RETRY = (
-        defer.TimeoutError,
-        TimeoutError,
-        DNSLookupError,
-        ConnectionRefusedError,
-        ConnectionDone,
-        ConnectError,
-        ConnectionLost,
-        TCPTimedOutError,
-        ResponseFailed,
-        IOError,
-        TunnelError,
-    )
-
-    def __init__(self, settings):
+    def __init__(self, settings: BaseSettings):
         if not settings.getbool("RETRY_ENABLED"):
             raise NotConfigured
         self.max_retry_times = settings.getint("RETRY_TIMES")
-        self.retry_http_codes = set(
-            int(x) for x in settings.getlist("RETRY_HTTP_CODES")
-        )
+        self.retry_http_codes = {int(x) for x in settings.getlist("RETRY_HTTP_CODES")}
         self.priority_adjust = settings.getint("RETRY_PRIORITY_ADJUST")
+        self.exceptions_to_retry = tuple(
+            load_object(x) if isinstance(x, str) else x
+            for x in settings.getlist("RETRY_EXCEPTIONS")
+        )
 
     @classmethod
-    def from_crawler(cls, crawler):
+    def from_crawler(cls, crawler: Crawler) -> Self:
         return cls(crawler.settings)
 
-    def process_response(self, request, response, spider):
+    def process_response(
+        self, request: Request, response: Response, spider: Spider
+    ) -> Request | Response:
         if request.meta.get("dont_retry", False):
             return response
         if response.status in self.retry_http_codes:
@@ -159,13 +148,21 @@ class RetryMiddleware:
             return self._retry(request, reason, spider) or response
         return response
 
-    def process_exception(self, request, exception, spider):
-        if isinstance(exception, self.EXCEPTIONS_TO_RETRY) and not request.meta.get(
+    def process_exception(
+        self, request: Request, exception: Exception, spider: Spider
+    ) -> Request | Response | None:
+        if isinstance(exception, self.exceptions_to_retry) and not request.meta.get(
             "dont_retry", False
         ):
             return self._retry(request, exception, spider)
+        return None
 
-    def _retry(self, request, reason, spider):
+    def _retry(
+        self,
+        request: Request,
+        reason: str | Exception | type[Exception],
+        spider: Spider,
+    ) -> Request | None:
         max_retry_times = request.meta.get("max_retry_times", self.max_retry_times)
         priority_adjust = request.meta.get("priority_adjust", self.priority_adjust)
         return get_retry_request(
